module EquationOfState


using dg1d
import Parameters: @unpack


export AbstractEquationOfState, ColdEquationOfState, HotEquationOfState,
       Polytrope, IdealGas, HybridEoS,
       Derivative, Density, Pressure, Enthalpy, TotalEnergy, InternalEnergy, Entropy,
       SpeedOfSound, SquareSpeedOfSound


# TODO Replace AbstractEquationOfState by AbstractEoS
# TODO Replace HotEquationOfState by HotEoS
# TODO Replace ColdEquationOfState by ColdEoS
# TODO Replace TVariable by TVar


#######################################################################
#                                Types                                #
#######################################################################


"""
    AbstractEquationOfState

Supertype for all Equations of State. To be used with the following method
definitions for the computation of thermodynamic variables `<:TVariable`:

```julia
function (eos::AbstractEquationOfState)(
    ::Type{TVariable_to_compute},
    ::Type{TVariable_1}, val1, ::Type{TVariable_2}, val2)

  # compute TVariable_to_compute from val1, val2
end
```

E.g. all equations of state are assumed to depend on two parameters.
Allowed pairs of argument types `(TVariable_1, TVariable_2)` are
- `(Density, InternalEnergy)`,
- `(Density, Pressure)`,
- `(InternalEnergy, Pressure)`,
and always in alphabetic order.

# Examples

```jldoctest setup
julia> eos = IdealGas(gamma=1.4);

julia> rho = eos(Density, InternalEnergy, 1.0, Pressure, 1.0)
2.5

julia> e = eos(TotalEnergy, Density, 1.0, InternalEnergy, 1.0)
2.0
```
"""
abstract type AbstractEquationOfState end

abstract type ColdEquationOfState <: AbstractEquationOfState end
abstract type HotEquationOfState  <: AbstractEquationOfState end


"""
    TVariable

Supertype for all theromdynamic variables.
TVariables participate in the `EquationOfState` interface.
"""
abstract type TVariable end


Base.broadcastable(eos::AbstractEquationOfState) = Ref(eos)
Base.broadcastable(var::TVariable) = Ref(var)


#######################################################################
#                  Available thermodynamic variables                  #
#######################################################################


"""
Denoted by p.
"""
struct Pressure       <: TVariable end


"""
Denoted by ρ.
"""
struct Density        <: TVariable end


"""
Per unit mass, denoted by ϵ.
"""
struct InternalEnergy <: TVariable end


"""
Per unit mass, denoted by s.
"""
struct Entropy        <: TVariable end


"""
    Derivative{T<:TVariable,S<:TVariable} <: TVariable

TVariable for requesting the first derivative `dT/dS`.

===

# Examples

```jldoctest setup
julia> eos = IdealGas(gamma=1.4);

julia> drhodp = eos(Derivative{Pressure,Density}, Density, 1.0, Pressure, 2.0)
0.8
```

See also: TVariable
"""
struct Derivative{T<:TVariable, S<:TVariable} <: TVariable end


#### Derived variables


"""
Per unit mass, denoted by e and defined as

  e = ρ (1 + ϵ).

"""
struct TotalEnergy    <: TVariable end


"""
Per unit mass, denoted by h.

  h = 1 + ϵ + p / ρ
"""
struct Enthalpy       <: TVariable end


"""
Denoted by cs.

  cs = sqrt((χ + (p / ρ^2) κ) / h)
"""
struct SpeedOfSound   <: TVariable end


"""
Denoted by cs^2.

  cs^2 = (χ + (p / ρ^2) κ) / h
"""
struct SquareSpeedOfSound   <: TVariable end



#######################################################################
#                          Interface methods                          #
#######################################################################


#### HotEquationOfState

# make eos arguments symmetric
# neat trick from https://discourse.julialang.org/t/is-multiple-dispatch-the-same-as-function-overloading/4145/5?u=fatteneder
(eos::HotEquationOfState)(TR, T1::Type{TT1}, a, T2::Type{TT2}, b) where {TT1<:TVariable,TT2<:TVariable} = eos(TR, T2, b, T1, a)


@nospecialize
# be specific about number of arguments
(eos::HotEquationOfState)(args...) = error("$(typeof(eos)) must be called with 5 arguments")

# fallback to not implemented error
# need two methods to disambig with symmetric definition above
(eos::HotEquationOfState)(TR::Type{<:TVariable}, T1::Type{<:TVariable}, a, T2,            b)                                       = error("$TR($T1,$T2) for $(typeof(eos)) not implemented")
(eos::HotEquationOfState)(TR::Type{<:TVariable}, T1::Type{TT1},         a, T2::Type{TT2}, b) where {TT1<:TVariable,TT2<:TVariable} = error("$TR($T1,$T2) for $(typeof(eos)) not implemented")

# error on duplicated argument types
(eos::HotEquationOfState)(TR::Type{<:TVariable}, T1::Type{T}, a, T2::Type{T}, b) where T<:TVariable = error("duplicated argument type found: $T")
(eos::HotEquationOfState)(TR::Type{T}, T1::Type{T}, a, T2, b) where T<:TVariable                    = error("duplicated argument type found: $T")
(eos::HotEquationOfState)(TR::Type{T}, T1::Type{T}, a, T2::Type{T}, b) where T<:TVariable           = error("duplicated argument type found: $T")
@specialize


#### ColdEquationOfState


@nospecialize
# be specific about number of arguments
(eos::ColdEquationOfState)(args...) = error("$(typeof(eos)) must be called with 3 arguments")

# fallback to not implement error
(eos::ColdEquationOfState)(TR::Type{TTR}, T1::Type{TT1}, a) where {TTR<:TVariable,TT1<:TVariable}= error("$TR($T1) for $(typeof(eos)) not implemented")

# error on duplicated argument types
(eos::ColdEquationOfState)(TR::Type{T}, T1::Type{T}, a) where T<:TVariable = error("duplicated argument type found: $T")
@specialize


# # for backwards compatibility
# # TODO Can we deprecate these?
# (eos::ColdEquationOfState)(::Type{Pressure}, ::Type{Density}, rho, ::Type{InternalEnergy}, eps)                            = eos(Pressure, Density, rho)
# (eos::ColdEquationOfState)(::Type{Density}, ::Type{InternalEnergy}, eps, ::Type{Pressure}, p)                              = eos(Density, Pressure, p)
# (eos::ColdEquationOfState)(::Type{Enthalpy}, ::Type{Density}, rho, ::Type{Pressure}, p)                                    = eos(Enthalpy, Density, rho)
# (eos::ColdEquationOfState)(::Type{InternalEnergy}, ::Type{Density}, rho, ::Type{Pressure}, p)                              = eos(InternalEnergy, Pressure, p)
# (eos::ColdEquationOfState)(::Type{Derivative{Pressure,Density}}, ::Type{Density}, rho, ::Type{InternalEnergy}, eps)        = eos(Derivative{Pressure,Density}, Density, rho)
# (eos::ColdEquationOfState)(::Type{Derivative{Pressure,InternalEnergy}}, ::Type{Density}, rho, ::Type{InternalEnergy}, eps) = eos(Derivative{Pressure,InternalEnergy}, Density, rho)


#######################################################################
#                           Implementations                           #
#######################################################################


include("derived_variables.jl")
include("polytrope.jl")
include("idealgas.jl")
include("hybrideos.jl")


function make_EquationOfState(prms)

  @unpack eos = prms
  # TODO Rename hybrideos to hybrid and fall back to concretetypes to figure
  # out available EoS implementations

  if eos == "hybrid"
    @unpack hybrid_cold_eos, hybrid_gamma = prms
    cold_eos = if hybrid_cold_eos == "polytrope"
      @unpack polytrope_k, polytrope_gamma = prms
      Polytrope(polytrope_k, polytrope_gamma)
    else
      error("EquationOfState: Unknown cold equation of state requested: $hybrid_eos")
    end
    return HybridEoS(cold_eos, hybrid_gamma)
  elseif eos == "idealgas"
    @unpack idealgas_gamma = prms
    return IdealGas(idealgas_gamma)
  elseif eos == "polytrope"
    @unpack polytrope_k, polytrope_gamma = prms
    return Polytrope(polytrope_k, polytrope_gamma)
  else
    error("EquationOfState: Unknown equation of state requested: $eos")
  end

end


@parameters EquationOfState begin

  """
  Available options
  - `polytrope`
  - `hybrid`
  - `idealgas`
  """
  eos = "polytrope"
  @check eos in [ "polytrope", "hybrid", "idealgas" ]

  "Polytropic constant"
  polytrope_k = 100.0
  @check polytrope_k > 0.0

  "exponent constant"
  polytrope_gamma = 2.0
  @check polytrope_gamma > 0.0

  "Thermalization constant"
  hybrid_gamma = 1.4
  @check hybrid_gamma > 0.0

  """
  Available options
  - `polytrope` ... adjust with `polytrope_k, polytrope_gamma`
  """
  hybrid_cold_eos = "polytrope"
  @check hybrid_cold_eos in [ "polytrope" ]

  "Polytropic constant"
  idealgas_gamma = 1.4
  @check idealgas_gamma > 0.0

end


end
