#######################################################################
#                              Timestep                               #
#######################################################################

timestep(env, P, hrsc) = timestep(env, P, hrsc, env.mesh)


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P
  @unpack max_v = get_static_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(max_v)
  # vmax > 0.99 && error("Unreasonable maximum speed encountered: $vmax")
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N^2 * vmax)
  return dt
end


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  max_v = get_variable(cache, :max_v)
  vmax = dg1d.absolute_maximum(view(max_v,:))
  vmax_limit = 0.99
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                                 2d                                  #
#######################################################################


function timestep(env, P, hrsc::Maybe{HRSC.AbstractHRSC}, mesh::Mesh2d)
  @unpack cache    = env
  @unpack equation = P
  @unpack max_v = get_static_variables(cache)

  broadcast_volume_2!(cons2prim, equation, mesh)
  broadcast_volume_2!(maxspeed, equation, mesh)
  vmax = dg1d.absolute_maximum(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack CFL, element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = CFL * dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


rhs!(env, P, hrsc, bdryconds, ldg_bdryconds, av_bdryconds) =
  rhs!(env, P, hrsc, bdryconds, ldg_bdryconds, av_bdryconds, env.mesh)

function rhs!(env, P::Project, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds, mesh::Mesh1d)

  @unpack cache, mesh              = env
  @unpack equation, rsolver        = P
  @unpack D, S, tau                = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau    = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau    = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau = get_bdry_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractReconstruction,
    bdryconds, ldg_bdryconds, av_bdryconds, ::Mesh1d)

  @unpack cache, mesh              = env
  @unpack equation, rsolver        = P
  @unpack D, S, tau                = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau    = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau    = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau = get_bdry_variables(cache)
  @unpack flag                     = get_cell_variables(cache)

  HRSC.reconstruct!(D,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(S,    flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(tau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D,    nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,    nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau,  nflx_tau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds, ::Mesh1d)

  @unpack cache, mesh                       = env
  @unpack equation, rsolver, ldg_rsolver,
          av_rsolver                        = P
  @unpack D, S, tau                         = get_dynamic_variables(cache)
  @unpack flx_D, flx_S, flx_tau,
          ldg_D, ldg_S, ldg_tau,
          flx_ldg_D, flx_ldg_S, flx_ldg_tau = get_static_variables(cache)
  @unpack rhs_D, rhs_S, rhs_tau             = get_rhs_variables(cache)
  @unpack nflx_D, nflx_S, nflx_tau, nflx_ldg_D,
          nflx_ldg_S, nflx_ldg_tau          = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_D,   flx_ldg_D,   nflx_ldg_D,   mesh)
  compute_rhs_weak_form!(ldg_S,   flx_ldg_S,   nflx_ldg_S,   mesh)
  compute_rhs_weak_form!(ldg_tau, flx_ldg_tau, nflx_ldg_tau, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  @. flx_D          += flx_ldg_D
  @. flx_S          += flx_ldg_S
  @. flx_tau        += flx_ldg_tau
  @. nflx_D   += nflx_ldg_D
  @. nflx_S   += nflx_ldg_S
  @. nflx_tau += nflx_ldg_tau

  compute_rhs_weak_form!(rhs_D,   flx_D,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_S,   flx_S,   nflx_S,   mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau, nflx_tau, mesh)

  return
end


#######################################################################
#                                 2d                                  #
#######################################################################


function rhs!(env, P, ::Mesh2d)

  @unpack cache, mesh                              = env
  @unpack equation                                 = P
  @unpack D, Sx, Sy, tau                           = get_dynamic_variables(cache)
  @unpack flx_D_x, flx_D_y, flx_Sx_x, flx_Sx_y,
          flx_Sy_x, flx_Sy_y, flx_tau_x, flx_tau_y = get_static_variables(cache)
  @unpack rhs_D, rhs_Sx, rhs_Sy, rhs_tau           = get_rhs_variables(cache)
  @unpack nflx_D, nflx_Sx, nflx_Sy, nflx_tau       = get_bdry_variables(cache)

  broadcast_volume_2!(cons2prim, equation, mesh)
  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_D, equation, mesh)
  broadcast_faces_2!(llf_Sx, equation, mesh)
  broadcast_faces_2!(llf_Sy, equation, mesh)
  broadcast_faces_2!(llf_tau, equation, mesh)

  compute_rhs_weak_form!(rhs_D,   flx_D_x,    flx_D_y,   nflx_D,   mesh)
  compute_rhs_weak_form!(rhs_Sx,  flx_Sx_x,   flx_Sx_y,  nflx_Sx,  mesh)
  compute_rhs_weak_form!(rhs_Sy,  flx_Sy_x,   flx_Sy_y,  nflx_Sy,  mesh)
  compute_rhs_weak_form!(rhs_tau, flx_tau_x,  flx_tau_y, nflx_tau, mesh)

  return
end
