#######################################################################
#                              Interface                              #
#######################################################################


"""
    abstract type AbstractProject end

The supertype for the Project interface of `dg1d`.

Subtypes must implement the constructor
```
function (::Type{<:AbstractProject})(env::Environment, prms)
  # ...
end
```

---

# Project interface

How to add a project called `MyProject`:

0. Setup a submodule in `src/projects/MyProject/MyProject.jl` called `MyProject`.
1. `include` the project in `src/projects/projects.jl`.
2. Define a subtype of `AbstractProject`. This will serve as the project's dispatcher.
3. Overload the following methods:
    - `dg1d.project_type`
    - `dg1d.parameters`
    - `dg1d.check_parameters`
    - `dg1d.rhs!`
    - `dg1d.timestep!`

---

# Example

```
# src/project/MyProject/MyProject.jl

module MyProject

using dg1d

# subtype AbstractProject
struct Project <: dg1d.AbstractProject
  # ...
end

function MyProject.Project(env::Environment, prms)
  # This constructor should
  # - register variables in cache(env)
  # - set up initial data for this project
  # - set up boundary conditions for this project
  # - set up callbacks for this project
end

dg1d.project_type(::Val{:MyModule}) = MyProject.Project

dg1d.@parameters MyModule begin
  ...
end

function dg1d.rhs!(env::dg1d.Environment, p::MyProject.Project)
  # compute the rhs inplace in `env`
end

function dg1d.timestep!(env::dg1d.Environment, p::MyProject.Project)
  # compute a timestep based on the state of `env`
end

end # MyModule
```
"""
abstract type AbstractProject end

function (::Type{T})(env::Environment, prms) where T<:AbstractProject
  error("Interface: Project type '$T' does not implement a constructor!")
end


"""
    project_type(::Val{T}) where T -> Type{S} where S<:AbstractProject

Register a module `T` with a dispatch type `S<:AbstractProject` in the dg1d
interface. `T` will also be the name for the parameter section in parameter files.
"""
function project_type(::Val{T}) where T
  error("Interface: Project '$T' does not provide a dispatch type")
end


"""
    parameters(::Val{T}) where T -> Dict{String,Any}

Return a dictionary with the default parameters for a project `T`.

Do not directly overload this method. Instead use `dg1d.@parameters` to combine
docstring generation with parameter definitions.

Parameters defined here can be modified from parameter files.
E.g. if project `MyProject` defines a parameter called `name`, you can modify it by
```
# parameter_file.toml
[MyProject]
name = 1
```

See: @parameters
---

# List of all parameters and their defaults available for inclusion in parameter files.

"""
function parameters(::Val{T}) where T
  error("Interface: Project '$T' does not provide default parameters")
end
parameters(s::Symbol) = parameters(Val(s))


"""
    check_parameters(::Val{T}, prms) where T -> nothing

Check consistency of parameters from `parameters(:MyProject)` and throw and
error if any check is violated.

Do not directly overload this method. Instead use `dg1d.@parameters` to combine
docstring generation with parameter definitions and optional checks.

See: @parameters
"""
function check_parameters(::Val{T}, prms) where T end
check_parameters(s::Symbol, prms) = check_parameters(Val(s), prms)


"""
    rhs!(env::Environment, p::AbstractProject)

Compute the rhs for project `p` inplace in `cache(env)`.
"""
function rhs!(env::Environment, p::AbstractProject, bdryconds)
  # TODO Remove bdryconds from this interface!!!
  error("Interface: Project type '$(typeof(p))' does not implement dg1d.rhs!")
end


# TODO Should a timestep computation be allowed to modify `cache(env)`?
"""
    timestep!(env::Environment, p::AbstractProject) -> Float64

Return a time step for project `p`.

The function is marked as inplace (note the `!`), because time step computations are allowed
to modify `cache(env)`.
"""
function timestep!(env::Environment, p::AbstractProject)
  error("Interface: Project type '$(typeof(p))' does not implement dg1d.timestep!")
end


function plot(env::Environment, p::AbstractProject)
  nothing
end


#######################################################################
#                               Helpers                               #
#######################################################################


function isregistered(p::Symbol)
  try
    project_type(Val(p))
    return true
  catch
    return false
  end
end
